"""
This module defines basic constants and functions that will be used
throughout nspyre

Author: Jacob Feder
Date: 7/8/2020
"""

###########################
# imports
###########################

# std
import os
from pathlib import Path

# 3rd party
from pint import UnitRegistry

###########################
# fundamental operations
###########################

# create a pint registry universal to nspyre
ureg = UnitRegistry()
Q_ = ureg.Quantity

# root directory of nspyre
NSPYRE_ROOT = Path(__file__).parent

def join_nspyre_path(path):
    """Return a full path from a path given relative to the nspyre root 
    directory"""
    return NSPYRE_ROOT / path

###########################
# resources
###########################

# config files
CLIENT_META_CONFIG_PATH = join_nspyre_path('config/client_meta_config.yaml')
SERVER_META_CONFIG_PATH = join_nspyre_path('config/server_meta_config.yaml')

# images
LOGO_PATH = str(join_nspyre_path('gui/images/spyre.png'))

###########################
# globals
###########################

# format for accessing instrument server devices from a client, where the first
# item is the server name and the second is the device name (both defined in the
# server config file)
INSERV_DEV_ACCESSOR = '{}/{}'
# mongodb replicaset name
MONGO_RS = 'NSpyreSet'
# all spyrelet databases in mongodb will be of this form
MONGO_SPYRELETS_KEY = 'spyrelet[{}]'
# mongodb instrument server databases will contain a special document
# that contains the instrument server settings
MONGO_SERVERS_SETTINGS_KEY = '_settings'
# All instrument server databases in mongodb will be of this form
MONGO_SERVERS_KEY = 'inserv[{}]'
# in ms
MONGO_CONNECT_TIMEOUT = 5000
# rpyc connection timeout in s
RPYC_CONN_TIMEOUT = None
# rpyc send/receive timeout in s
RPYC_SYNC_TIMEOUT = None

# config file key for mongodb address
CONFIG_MONGO_ADDR_KEY = 'mongodb_addr'



import itertools as it

from PyQt5 import QtWidgets, QtCore

import pyqtgraph as pg
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph import functions as fn

pg.setConfigOptions(imageAxisOrder='row-major')
import numpy as np

from nspyre.gui.colormap import viridis
from nspyre.gui.colors import cyclic_colors, colors
from nspyre.gui.widgets.splitter_widget import Splitter, SplitterOrientation

from nspyre.gui.widgets.spinbox import SpinBox
from nspyre.gui.widgets.code_editor import Scintilla_Code_Editor, Monokai_Python_Lexer

import traceback
import inspect
from scipy.optimize import curve_fit

from collections import OrderedDict


class BasePlotWidget(QtWidgets.QWidget):

    def __init__(self, w=None, plot_item=None, parent=None):
        super().__init__(parent=parent)
        if w is None:
            w = pg.PlotWidget()
        if plot_item is None:
            plot_item = w.getPlotItem()
        self.w = w
        self.plot_item = plot_item
        self.xonbottom = True
        self.yonleft = True
        self.traces = dict()

        self._title = ''
        self._xlabel = ''
        self._ylabel = ''
        self.invertY = False
        self.invertX = False

        self.init_plot()

        return

    def init_plot(self):
        self.build_toolbox()

        splitter_config = {
            'main_w': self.w,
            'side_w': self.toolbox,
            'orientation': SplitterOrientation.vertical_left_button,
        }
        splitter = Splitter(**splitter_config)

        def show_event(ev):
            h = splitter.size().height()
            splitter.setSizes([1, 0])
            splitter.resize(h, h)

        # Collaspe the tools to start
        splitter.showEvent = show_event

        layout = QtWidgets.QGridLayout()
        layout.addWidget(splitter)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)
        return

    def build_toolbox(self):
        # create default tools
        self.crosshairs = CrosshairAddon(self.plot_item)

        # create toolbox and add tools
        self.toolbox = QtWidgets.QToolBox()
        self.toolbox.addItem(self.crosshairs, "Crosshairs")
        return

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, _title):
        self._title = _title
        self.plot_item.setTitle(title=_title)
        return

    @property
    def xlabel(self):
        return self._xlabel

    @xlabel.setter
    def xlabel(self, _xlabel):
        self._xlabel = _xlabel
        pos = 'bottom' if self.xonbottom else 'top'
        self.plot_item.setLabel(pos, _xlabel)
        return

    @property
    def ylabel(self):
        return self._ylabel

    @ylabel.setter
    def ylabel(self, _ylabel):
        self._ylabel = _ylabel
        pos = 'left' if self.yonleft else 'right'
        self.plot_item.setLabel(pos, _ylabel)
        return

    def clear(self):
        ypos = 'left' if self.yonleft else 'right'
        xpos = 'bottom' if self.xonbottom else 'top'
        self.plot_item.setLabels(**{'title':'', xpos:'', ypos:''})

    @property
    def invertY(self):
        return self._invertY

    @invertY.setter
    def invertY(self, invert):
        self.plot_item.getViewBox().invertY(invert)
        self._invertY = invert
        return

    @property
    def invertX(self):
        return self._invertX

    @invertX.setter
    def invertX(self, invert):
        self.plot_item.getViewBox().invertX(invert)
        self._invertX = invert
        return

    def generate_meta(self, **kwargs):
        d = {'xlabel':self._xlabel, 'ylabel':self._ylabel, 'title':self._title, 'invertX':self._invertX, 'invertY':self._invertY}
        d.update(kwargs)
        return d


    def load_meta(self, meta, params=['xlabel', 'ylabel', 'title', 'invertX', 'invertY']):
        for key in params:
            if key in meta and hasattr(self, key):
                setattr(self, key, meta[key])

class HeatmapPlotWidget(BasePlotWidget):

    def __init__(self, parent=None, cmap=None):
        plot_item = pg.PlotItem(enableMouse=False)
        w = pg.ImageView(view=plot_item)
        super().__init__(parent=parent, w=w, plot_item=plot_item)
        w.ui.roiBtn.clicked.connect(self._set_roi_pos)
        self.grid(False)
        gradient = self.w.ui.histogram.gradient
        if cmap is None:
            cmap = viridis
        gradient.setColorMap(cmap)
        for tick in gradient.ticks:
            tick.hide()
        gradient.setFixedWidth(300)
        # disable gradient editing
        gradient.mouseClickEvent = lambda ev: None
        self.add_plotting_options()
        self._pos = None
        self._scale = None

        self.invertY = True
        return

    def _set_roi_pos(self):
        #Tries to place the ROI in a good spot when clicking the button
        r = np.array(self.plot_item.getViewBox().viewRange())
        pos = r.mean(axis=1)
        size = np.diff(r, axis=1)[:,0]
        if any(abs(pos-self.w.roi.pos())>abs(size)):
            self.w.roi.setPos(pos)
        if any(2*size<self.w.roi.size()):
            self.w.roi.setSize(size/2)


    def add_plotting_options(self):
        layout = QtWidgets.QVBoxLayout()
        self.plot_opts_checkboxes = dict()
        for k,val in [('autoHistogramRange',False), ('autoLevels',True), ('autoRange',True)]:
            self.plot_opts_checkboxes[k] = w = QtWidgets.QCheckBox(k)
            w.setChecked(val)
            layout.addWidget(w)
        
        #Add aspect lock option
        self.plot_opts_checkboxes['aspectLocked'] = w = QtWidgets.QCheckBox('aspectLocked')
        w.setChecked(True)
        layout.addWidget(w)
        w.stateChanged.connect(lambda x: setattr(self, 'aspectLocked', self.plot_opts_checkboxes['aspectLocked'].isChecked()))


        tool_w = QtWidgets.QWidget()
        tool_w.setLayout(layout)
        self.toolbox.addItem(tool_w, "Plotting Options")


    def grid(self, toggle, alpha=0.4):
        self.plot_item.showGrid(x=toggle, y=toggle, alpha=alpha)
        return

    @property
    def aspectLocked(self):
        return self.w.getView().state['aspectLocked']

    @aspectLocked.setter
    def aspectLocked(self, lock):
        return self.w.getView().setAspectLocked(lock)

    @property
    def im_pos(self):
        return self._pos

    @im_pos.setter
    def im_pos(self, pos):
        self._pos = pos

    @property
    def im_scale(self):
        return self._scale

    @im_scale.setter
    def im_scale(self, scale):
        self._scale = scale

    def set(self, im):
        self.w.setImage(im, pos=self._pos, scale=self._scale,
                        autoRange=self.plot_opts_checkboxes['autoRange'].isChecked(),
                        autoLevels=self.plot_opts_checkboxes['autoLevels'].isChecked(),
                        autoHistogramRange=self.plot_opts_checkboxes['autoHistogramRange'].isChecked(),)
        return

    def get(self):
        return self.w.getImageItem().image


class LinePlotWidget(BasePlotWidget):

    plots_updated = QtCore.pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.legend = self.w.addLegend()
        self.grid(True)
        self._colors = it.cycle(cyclic_colors)
        self.install_fitter()
        return

    def install_fitter(self):
        self.fitter = FitterWidget(self.w)
        self.fitter.traces = self.traces
        self.toolbox.addItem(self.fitter, "Fitter")
        return

    def grid(self, toggle, alpha=0.4):
        self.w.showGrid(x=toggle, y=toggle, alpha=alpha)
        return

    def plot(self, tracename, **kwargs):
        symbol_pen = kwargs.get('symbol_pen', pg.mkPen(color=(255, 255, 255, 100)))
        symbol_brush = kwargs.get('symbol_brush', pg.mkBrush(color=(255, 255, 255, 100)))
        symbol_size = kwargs.get('symbol_size', 5)
        symbol = kwargs.get('symbol', 's')
        pen = kwargs.get('pen', pg.mkPen(color=(next(self._colors) + (200,)), width=1))
        antialias = kwargs.get('antialias', False)
        trace = self.w.plot(
            symbol=symbol,
            symbolSize=symbol_size,
            symbolPen=symbol_pen,
            symbolBrush=symbol_brush,
            pen=pen,
            antialias=antialias,
        )
        trace.curve.setClickable(True)
        trace_err = None
        self.legend.addItem(trace, tracename)
        self.traces[tracename] = trace, trace_err
        self.fitter.update_traces()
        return

    def remove_trace(self, tracename):
        for item in self.traces.pop(tracename):
            self.plot_item.removeItem(item)
        self.legend.removeItem(tracename)
        self.fitter.update_traces()

    def clear(self):
        super().clear()
        for tracename in list(self.traces.keys()):
            self.remove_trace(tracename)
        self._colors = it.cycle(cyclic_colors)

    def set(self, tracename, **kwargs):
        if tracename not in self.traces:
            self.plot(tracename, **kwargs)
        trace, trace_err = self.traces[tracename]
        data = kwargs.pop('data', None)
        xs = kwargs.pop('xs', None)
        ys = kwargs.pop('ys', None)
        yerrs = kwargs.pop('yerrs', None)
        if not any(item is None for item in [data, xs, ys]):
            raise ValueError('No plot points supplied (either data or xs and ys must be given)')
        if data is not None:
            xs, ys = list(zip(*data))
        if isinstance(xs, list):
            xs = np.array(xs)
        elif not isinstance(xs, np.ndarray):
            xs = xs.values
        if isinstance(ys, list):
            ys = np.array(ys)
        elif not isinstance(ys, np.ndarray):
            ys = ys.values
        trace.setData(x=xs, y=ys, **kwargs)
        if yerrs is not None:
            if trace_err is None:
                error_bar_params = {
                    'x': xs,
                    'y': ys,
                    'top': 0,
                    'bottom': 0,
                    'beam': 0.0,
                    'pen': pg.mkPen(color=(255, 255, 255, 80)),
                }
                trace_err = pg.ErrorBarItem(**error_bar_params)
                trace_err.setZValue(-999)
                self.w.addItem(trace_err)
                self.traces[tracename] = trace, trace_err
            ylines = list(zip(*[(yerr / 2, yerr / 2) if not np.isnan(yerr) else (0, 0) for y, yerr in zip(ys, yerrs)]))
            if ylines:
                ybottoms, ytops = ylines
                trace_err.setData(x=xs, y=ys, top=ytops, bottom=ybottoms, beam=0.0)
        return

    def get(self, tracename):
        pditem, _ = self.traces[tracename]
        x, y = pditem.getData()
        if x is None or y is None:
            return None, None
        else:
            return x.astype(np.float), y.astype(np.float)

    def __iter__(self):
        for tracename in self.traces:
            yield tracename

    def __getitem__(self, tracename):
        return self.get(tracename)



class FastImageWidget(BasePlotWidget):

    def __init__(self, parent=None):
        graphic_view = pg.GraphicsView()
        plot_item = pg.PlotItem(enableMouse=False)
        self.img_item = pg.ImageItem()
        plot_item.addItem(self.img_item)
        graphic_view.setCentralItem(plot_item)
        super().__init__(parent=parent, w=graphic_view, plot_item=plot_item)

    def set(self, image=None, autoLevels=None, **kargs):
        self.img_item.setImage(image=image, autoLevels=autoLevels, **kargs)


##---------------------------------------------------------------
##                      Crosshairs
##---------------------------------------------------------------
class Crosshair(QtCore.QObject):
    sigPositionChanged = QtCore.pyqtSignal(object)
    sigPositionChangeFinished = QtCore.pyqtSignal(object)

    def __init__(self, plot_item, pos, **kwargs):
        super().__init__()
        self.kwargs = kwargs
        self.moving = False
        self.hovering = False

        pen = fn.mkPen((255, 255, 0, 127))
        self.vLine = pg.InfiniteLine(angle=90, pen=pen, movable=False)
        self.hLine = pg.InfiniteLine(angle=0,  pen=pen, movable=False)
        self.vLine.hoverEvent, self.hLine.hoverEvent = self.hoverEvent, self.hoverEvent
        self.vLine.mouseDragEvent, self.hLine.mouseDragEvent = self.mouseDragEvent, self.mouseDragEvent

        self.center_dot = pg.ScatterPlotItem(pos=[pos], pen=fn.mkPen((255,0,0, 127)), brush=(255,0,0), symbol='o', size=3)

        plot_item.addItem(self.vLine, ignoreBounds=True)
        plot_item.addItem(self.hLine, ignoreBounds=True)
        plot_item.addItem(self.center_dot, ignoreBounds=True)
        self.plot_item = plot_item
        self.set_pos(pos)

    def set_pos(self, pos, emit_sig=True):
        if isinstance(pos, QtCore.QPointF):
            self.pos = [pos.x(), pos.y()]
        else:
            self.pos = list(pos)
        self.vLine.setPos(self.pos[0])
        self.hLine.setPos(self.pos[1])
        self.center_dot.setData(pos=[pos])
        if emit_sig: self.sigPositionChanged.emit(self.get_pos())

    def mouseDragEvent(self, ev):
        if ev.button() == QtCore.Qt.LeftButton:
            if ev.isStart():
                self.moving = True
            ev.accept()

            if not self.moving:
                return
            self.set_pos(self.plot_item.vb.mapSceneToView(ev.scenePos()))
            if ev.isFinish():
                self.moving = False
                self.sigPositionChangeFinished.emit(self.get_pos())

    def hoverEvent(self, ev):
        if (not ev.isExit()) and ev.acceptDrags(QtCore.Qt.LeftButton):
            self.hovering = True
            for line in [self.vLine, self.hLine]: line.currentPen = fn.mkPen(255, 0,0)
        else:
            self.hovering = False
            for line in [self.vLine, self.hLine]: line.currentPen = line.pen
        for line in [self.vLine, self.hLine]:
            line.update()

    def get_pos(self):
        return self.pos

    def delete(self):
        self.hLine.deleteLater()
        self.vLine.deleteLater()
        self.center_dot.deleteLater()


class CrosshairAddon(QtWidgets.QWidget):
    sigCrosshairAdded = QtCore.pyqtSignal(object)
    sigCrosshairRemoved = QtCore.pyqtSignal(object)
    def __init__(self, plot_item, **kwargs):
        super().__init__(**kwargs)
        self.plot_item = plot_item
        self.cross_list = list()
        self._spinbox_decimals = 4

        self.build_ui()


    @property
    def spinbox_decimals(self):
        return self._spinbox_decimals

    @spinbox_decimals.setter
    def spinbox_decimals(self, val):
        if self._spinbox_decimals != val:
            for r in range(self.table.rowCount()):
                self.table.cellWidget(r,0).setDecimals(val)
                self.table.cellWidget(r,1).setDecimals(val)
            self._spinbox_decimals = val

    def build_ui(self):
        self.table = QtWidgets.QTableWidget()
        self.table.setColumnCount(3)
        self.table.setHorizontalHeaderLabels(['x','y','Delete'])

        #Add control (for now just an add button)
        add_btn = QtWidgets.QPushButton('+ Add')
        def add():
            r = np.array(self.plot_item.getViewBox().viewRange())
            self.add_crosshair(r.mean(axis=1))
        add_btn.clicked.connect(lambda: add())

        #Add a decimal precision box
        decimal_input = SpinBox(value=self.spinbox_decimals, minStep=1, dec=False, int=True, bounds=(0, None), step=1)
        decimal_input.valueChanged.connect(lambda x: setattr(self, 'spinbox_decimals', decimal_input.value()))

        ctrl_layout = QtWidgets.QFormLayout()
        ctrl_layout.addRow('Add Crosshair', add_btn)
        ctrl_layout.addRow('Floating point precision', decimal_input)
        ctrl_widget = QtWidgets.QWidget()
        ctrl_widget.setLayout(ctrl_layout)


        layout = QtWidgets.QGridLayout()
        layout.addWidget(ctrl_widget)
        layout.addWidget(self.table)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def add_crosshair(self, pos, **kwargs):
        # Add the Crosshair to the list
        cross = Crosshair(self.plot_item, pos, **kwargs)
        self.cross_list.append(cross)

        # Add the table entry
        row = len(self.cross_list)-1
        self.table.insertRow(row)

        # Add the x,y widgets
        def update_pos(axis, value):
            if cross.moving:
                return
            cur = cross.get_pos()
            if axis==0:
                cross.set_pos([value, cur[1]], emit_sig=False)
            elif axis==1:
                cross.set_pos([cur[0], value], emit_sig=False)
        def lambda_gen(axis):
            return lambda obj: update_pos(axis, obj.value())

        for i in range(2):
            w = SpinBox(value = pos[i], dec=True, decimals=self.spinbox_decimals)
            self.table.setCellWidget(row, i, w)
            w.sigValueChanged.connect(lambda_gen(i))


        # Add a remove button
        btn = QtWidgets.QPushButton('X')
        btn.clicked.connect(lambda: self.remove_crosshair(cross))
        self.table.setCellWidget(row, 2, btn)

        # Link the position of the cross to the numbers in the table
        cross.sigPositionChanged.connect(lambda: self.update_table_entry(cross))
        self.sigCrosshairAdded.emit(cross)


    def _find_index(self, cross):
        for i in range(len(self.cross_list)):
            if self.cross_list[i] == cross:
                return i

    def update_table_entry(self, cross):
        row = self._find_index(cross)
        self.table.cellWidget(row, 0).setValue(cross.get_pos()[0])
        self.table.cellWidget(row, 1).setValue(cross.get_pos()[1])

    def remove_crosshair(self, cross):
        index = self._find_index(cross)
        self.table.removeRow(index)
        cross.delete()
        self.cross_list.pop(index)
        self.sigCrosshairRemoved.emit(index)


    def __getitem__(self, k):
        return self.cross_list[k].get_pos()

    def __iter__(self):
        for cross in self.cross_list:
            yield cross

    def __len__(self):
        return len(self.cross_list)

class FitterWidget(QtWidgets.QWidget):

    EXTRA_TEXT = """
# Fitter rules:
#     - Fit function name is "fit_func"
#     - Default values to arguments will be 
#       passed as start values for the fitter
#     - First argument of "fit_func" must be "xs"
#     - Have fun!

import numpy as np
"""

    DEFAULT_FUNCTIONS = {
        'laurentzian':"""
# Laurentzian
def fit_func(xs, A=1, gamma=1, x0=0, y0=0):
    return A/(1+(2*(xs-x0)/gamma)**2)+y0
    """,
        'gaussian':"""
# Gaussian
def fit_func(xs, A=1, sigma=1, x0=0, y0=0):
    return A*np.exp(-0.5*((xs-x0)/sigma)**2)+y0
    """,
        'cos':"""
# Cosine
def fit_func(xs, A=1, T=50, phi=0, y0=1):
    return A*np.cos(np.pi*xs/T+phi)+y0
    """,
        'linear':"""
# Linear
def fit_func(xs, m=1, y0=0):
    return m*xs + y0
    """,
    }

    def __init__(self, w, parent=None):
        super().__init__(parent=parent)
        self.w = w
        self.traces = dict()
        self.fits = dict()
        self.init_ui()
        return

    def init_ui(self):
        self.traces_list = QtWidgets.QComboBox()

        self.editor = Scintilla_Code_Editor()
        lexer = Monokai_Python_Lexer(self.editor)
        self.editor.setLexer(lexer)

        self.func_list = QtWidgets.QComboBox()
        self.func_list.addItems([func_name for func_name in sorted(self.DEFAULT_FUNCTIONS.keys())])
        def update_text(*args):
            selected_func = self.func_list.currentText()
            txt = self.EXTRA_TEXT + self.DEFAULT_FUNCTIONS[selected_func]
            self.editor.setText(txt)
        self.func_list.currentIndexChanged.connect(update_text)
        update_text()

        self.results_table = QtWidgets.QTableWidget()

        self.update_traces()
        self.fit = QtWidgets.QPushButton('Fit')
        self.remove = QtWidgets.QPushButton('Remove fit')
        self.remove_all = QtWidgets.QPushButton('Remove all fits')
        self.fit.clicked.connect(self.compile_and_fit)
        self.remove.clicked.connect(self.remove_fit)
        self.remove_all.clicked.connect(self.remove_all_fits)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.traces_list)
        layout.addWidget(self.func_list)
        layout.addWidget(self.editor)
        layout.addWidget(self.results_table)
        layout.addWidget(self.fit)
        layout.addWidget(self.remove)
        layout.addWidget(self.remove_all)
        self.setLayout(layout)
        return

    def update_results(self):
        selected_trace_name = self.traces_list.currentText()
        if selected_trace_name in self.fits:
            self.results_table.clear()
            self.results_table.setColumnCount(2)
            self.results_table.setHorizontalHeaderLabels(['param', 'value'])

            params = self.fits[selected_trace_name]['params']
            self.results_table.setRowCount(len(params))

            for i, name in enumerate(params.keys()):
                self.results_table.setItem(i,0, QtWidgets.QTableWidgetItem(name))
                self.results_table.setItem(i,1, QtWidgets.QTableWidgetItem(str(params[name])))

    def update_traces(self):
        self.traces_list.clear()
        self.traces_list.addItems([plot_name for plot_name in sorted(self.traces.keys())])
        return

    def update_fits(self):
        for trace_name, data in self.fits.items():
            xs = data['xs']
            ys = data['ys']
            curve = data['curve']
            if curve is None:
                curve = self.w.plot(pen=pg.mkPen(color=colors['yellow'], width=1), antialias=True)
                data['curve'] = curve
            curve.setData(x=xs, y=ys)
        return

    def remove_fit(self):
        selected_trace_name = self.traces_list.currentText()
        try:
            fit_data = self.fits[selected_trace_name]
            fit = fit_data['curve']
        except KeyError:
            return
        self.w.removeItem(fit)
        del self.fits[selected_trace_name]
        return

    def remove_all_fits(self):
        for name in (self.traces_list.itemText(idx) for idx in range(self.traces_list.count())):
            try:
                fit_data = self.fits[name]
                fit = fit_data['curve']
            except KeyError:
                continue
            self.w.removeItem(fit)
            del self.fits[name]
        return

    def compile_and_fit(self):
        selected_trace_name = self.traces_list.currentText()
        trace, _ = self.traces[selected_trace_name]
        plot_xs, plot_ys = trace.xData, trace.yData

        # Execute the code
        glob = globals()
        exec(self.editor.text(), glob)
        fit_func = glob['fit_func']

        # Get some default values
        sig = inspect.signature(fit_func)
        p0 = list()
        for i, p in enumerate(sig.parameters.values()):
            if i == 0:
                pass # To remove the xs parametter
            elif p.default is inspect._empty:
                p0.append(1)
            else:
                p0.append(p.default)

        # Perform the curve fitting
        popt, pcov = curve_fit(fit_func, plot_xs, plot_ys, p0=p0)
        fit_ys = fit_func(plot_xs, *popt)

        names = list(sig.parameters.keys())
        params = OrderedDict([(names[i+1], popt[i]) for i in range(len(popt))])

        self.fits[selected_trace_name] = {
            'xs': plot_xs,
            'ys': fit_ys,
            'curve': self.fits.get(selected_trace_name, dict()).get('curve'),
            'params': params,
            'pcov': pcov
        }
        self.update_fits()
        self.update_results()


#!/usr/bin/env python
"""
CLI for controlling mongodb

Author: Jacob Feder
Date: 9/13/2020
"""

###########################
# imports
###########################

# std
import argparse
from sys import platform
from pathlib import Path
import subprocess

###########################
# globals
###########################

THIS_DIR = Path(__file__).parent

###########################
# exceptions
###########################

class OSNotSupportedError(Exception):
    pass

###########################
# classes / functions
###########################

def main():
    """Entry point for mongodb CLI"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-mongodb',
                            description='Start / restart the MongoDB server')
    cmd_args = arg_parser.parse_args()
    if platform == 'linux' or platform == 'linux2' or platform == 'darwin':
        subprocess.run(['bash', str(THIS_DIR / 'start_mongo_unix.sh')])
    elif platform == 'win32':
        subprocess.run([str(THIS_DIR / 'start_mongo_win.bat')])
    else:
        raise OSNotSupportedError('Your OS [{}] is not supported'.\
                                    format(platform))

###########################
# standalone main
###########################

if __name__ == '__main__':
    main()



"""
    nspyre.spyrelet.spyrelet.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    This module defines the base functionality of spyrelets, which are scripts
    that encode the experimental logic.

    Author: Alexandre Bourassa
    Date: 10/30/2019
    Modified: Jacob Feder 7/11/2020
"""

###########################
# imports
###########################

# std
import threading
from collections import OrderedDict
import traceback
import inspect
from copy import copy
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
import sys
from importlib import import_module
import logging

# 3rd party
import pymongo
import pandas as pd
import numpy as np
from pint.util import infer_base_unit
from pymongo.errors import PyMongoError
from tqdm.auto import tqdm

# nspyre
from nspyre.gui.data_handling import save_data
from nspyre.config.config_files import load_config, get_config_param, \
                                    ConfigEntryNotFoundError
from nspyre.utils.misc import get_mongo_client, custom_decode, custom_encode, \
                                RangeDict, load_class_from_file
from nspyre.definitions import Q_

###########################
# globals
###########################

# config file key for spyrelets
CONFIG_SPYRELETS_KEY = 'spyrelets'
# config file key for the python file of the spyrelet
CONFIG_SPYRELETS_FILE_KEY = 'file'
# config file key for the class of the spyrelet
CONFIG_SPYRELETS_CLASS_KEY = 'class'
# config file key for the device aliases of a spyrelet
CONFIG_DEVS_KEY = 'device_aliases'
# config file key for the sub-spyrelets of a spyrelet
CONFIG_SUB_SPYRELETS_KEY = 'spyrelets'
# config file key for spyrelet arguments
CONFIG_SPYRELETS_ARGS_KEY = 'args'
# record of the loaded spyrelets
_LOADED_SPYRELETS = {}

###########################
# exceptions
###########################

class StopRunning(Exception):
    pass

class SpyreletLoadError(Exception):
    """Exception while loading a Spyrelet"""
    def __init__(self, error, msg):
        super().__init__(msg)
        if error:
            logging.exception(error)

class SpyreletUnloadError(Exception):
    """Exception while unloading a Spyrelet"""
    def __init__(self, error, msg):
        super().__init__(msg)
        if error:
            logging.exception(error)

###########################
# classes / functions
###########################

class Spyrelet():
    """
    A few notes about the spyrelet class:
        - This is the class you need to subclass for making experiments.
        - All devices used in the spyrelet must be listed in the
            REQUIRED_DEVICES dict
        - All sub-spyrelet must also be listed in the REQUIRED_SPYRELETS dict
        - Upon instantiation the class will check the __init__ arguments
            devices and spyrelets to make sure they satisfy these requirements
        - For higher performance we will store the data internally as a list
            instead of a dataframe.  Quicker to append to a list.
    """

    # A dict with the names and associated class of the devices required to run
    # this spyrelet
    REQUIRED_DEVICES = list()

    # A dict with the name and associated class of the sub-spyrelet required to
    # run this spyrelet
    REQUIRED_SPYRELETS = dict()

    # A definition of the parametters that are used as arguments to the
    # main/initialize/finalize functions. These are used both to generate a
    # launcher GUI and to enforce units at call time.
    PARAMS = dict()

    # An extra dictionary, which can be defined by the user at initialization
    # time. This can store anything the users want
    CONSTS = dict()

    def __init__(self, unique_name, gateway, device_aliases={}, spyrelets={},
                    mongodb_addr=None, **consts):
        self.name = unique_name
        self.progress = tqdm
        self.spyrelets = spyrelets
        self.CONSTS = self.CONSTS.copy()
        self.CONSTS.update(**consts)
        self.last_kwargs = dict()

        self.mongodb_addr = mongodb_addr
        self.validate()
        
        reg_entry = {
            '_id': unique_name,
            'class': 'unused' #"{}.{}".format(self.__class__.__module__, self.__class__.__name__),
        }

        self.client = get_mongo_client(mongodb_addr)
        self.col = self.client['Spyre_Live_Data'][unique_name]
        self.client['Spyre_Live_Data']['Register'].update_one({'_id':unique_name},{'$set':reg_entry}, upsert=True)
        self.clear_data()
        
        # check that the server devices are accessible and add them as
        # instance variables
        for dev_alias in self.REQUIRED_DEVICES:
            try:
                # e.g. "local1/dev1"
                dev_accessor = device_aliases[dev_alias]
            except Exception as exc:
                raise SpyreletLoadError(exc, 'Spyrelet [{}] requires the '
                        'device [{}] but it wasn\'t defined in the [{}] '
                        'section of the config file'.format(unique_name, \
                            dev_alias, CONFIG_DEVS_KEY)) from None
            try:
                server_name, device_name = device_aliases[dev_alias].split('/')
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] with the '
                    'device alias [{}] has an invalid device accessor [{}].'
                    'The accessor should be in the form '
                    '"server_name/device_name"'.format(unique_name, dev_accessor, dev_alias))
            try:
                server = getattr(gateway, server_name)
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] requires the '
                        'device [{}] (alias [{}]) but the instrument '
                        'server is unreachable'.format(unique_name, dev_accessor, dev_alias))
            try:
                device = getattr(server, device_name)
            except:
                raise SpyreletLoadError(None, 'Spyrelet [{}] requires the '
                        'device [{}] (alias [{}]) but the instrument '
                        'server doesn\'t contain the device'.format(unique_name, dev_accessor, dev_alias))
            setattr(self, dev_alias, device)

        # check that the sub spyrelets are loaded and add them as
        # instance variables
        for sname, sclass in self.REQUIRED_SPYRELETS.items():
            if sname in spyrelets:# and isinstance(spyrelets[sname], sclass):
                setattr(self, sname, spyrelets[sname])
            else:
                raise SpyreletLoadError('Spyrelet [{}] requires the '
                        'sub-spyrelet [{}] but it wasn\'t loaded'.format(unique_name, sname))

    def set_defaults(self, **params_dict):
        d = {'defaults.{}'.format(key):val for key,val in custom_encode(params_dict).items()}
        if len(d):
            return self.client['Spyre_Live_Data']['Register'].update_one({'_id':self.name},{'$set':d}, upsert=True)

    def run(self, *args, **kwargs):
        self.progress = kwargs.pop('progress') if ('progress' in kwargs) else tqdm
        clear_data = kwargs.pop('clear_data') if ('clear_data' in kwargs) else True
        if self.progress is None: self.progress = lambda *args, **kwargs: tqdm(*args, leave=False, **kwargs)
        try:
            args, kwargs = self.enforce_args_units(*args, **kwargs)
            self._stop_flag = False
            if clear_data:
                self.clear_data()
            self.initialize(*args, **kwargs)
            self.main(*args, **kwargs)
        except StopRunning:
            print('stopping spyrelet')
            pass
        except:
            traceback.print_exc()
        finally:
            self.finalize(*args, **kwargs)

    def bg_run(self, *args, **kwargs):
        t = threading.Thread(target=lambda: self.run(*args, **kwargs))
        t.start()
        return t

    def enforce_args_units(self, *args, **kwargs):
        args = list(args)
        def _enforce_units(val, param):
            if 'units' in param:
                if type(val) is Q_:
                    return val.to(param['units'])
                else:
                    return val*Q_(1, param['units'])
            else:
                return val

        sig = inspect.signature(self.main)
        params = list(sig.parameters.keys())
        for name in list(self.PARAMS.keys()):
            param_index = params.index(name)
            if param_index >= len(args):
                if name in kwargs:
                    kwargs[name] = _enforce_units(kwargs[name], self.PARAMS[name])
                else:
                    kwargs[name] = _enforce_units(sig.parameters[name].default, self.PARAMS[name])
            else:
                args[param_index] = _enforce_units(args[param_index], self.PARAMS[name])
        return args, kwargs

    def main(self, *args, **kwargs):
        """This is the method that will contain the user main logic. \
        Should be overwritten"""
        raise NotImplementedError

    def initialize(self, *args, **kwargs):
        """This is the method that will contain the user initialize logic. \
        Should be overwritten"""
        pass

    def finalize(self, *args, **kwargs):
        """This is the method that will contain the user finalize logic. \
        Should be overwritten. This will run even if the initialize or main \
        errors out.
        """
        pass

    def call(self, spyrelet, *args, **kwargs):
        """This is the proper way to call a sub spyrelet. It will take care of \
        keeping the data and calling the proper progress bar. If use_defaults \
        is True (defaults to True) every argument needs to be passed as a \
        keyword argument (ie *args will be ignored).
        """
        keep_data = kwargs.pop('keep_data') if 'keep_data' in kwargs else True
        # use_defaults = kwargs.pop('use_defaults') if 'use_defaults' in kwargs else True
        ignore_child_progress = kwargs.pop('ignore_child_progress') if 'ignore_child_progress' in kwargs else False

        if ignore_child_progress:
            progress = lambda x: x
        else:
            progress = self.progress

        # if not use_defaults:
        spyrelet.run(*args, progress=progress, **kwargs)
        # else:
        #     launcher = Spyrelet_Launcher(spyrelet)
        #     launcher.run(progress=self.progress,**kwargs)
        
        if keep_data:
            if not spyrelet.name in self._child_data:
                self._child_data[spyrelet.name] = list()
            self._child_data[spyrelet.name].append(spyrelet.data)


    def stop(self):
        self._stop_flag = True
        for sname in self.REQUIRED_SPYRELETS:
            getattr(self, sname).stop()

    def clear_data(self):
        self.col.drop()
        self._data = list()
        self._child_data = dict()

    CASTING_DICT = {np.int32:int, np.float64:float}
    def acquire(self, row):
        # Cleanup row
        # Here we will keep numpy arrays as is for local copy, but change it to list for MongoDB
        if not row is None:
            restore_row = dict()
            for k, val in row.items():
                if type(val) == Q_:
                    base_unit = '*'.join('{} ** {}'.format(u, p) for u, p in infer_base_unit(val).items())
                    row[k] = row[k].to(Q_(base_unit)).m
                if type(val) == np.ndarray:
                    restore_row[k] = row[k]
                    row[k] = row[k].tolist()
                if type(val) in self.CASTING_DICT:
                    row[k] = self.CASTING_DICT[type(val)](row[k])

            self.col.insert_one(row)

            for k, val in restore_row.items():
                row[k] = val
            self._data.append(row)
        if self._stop_flag:
            raise StopRunning

    """
    The cache allows for passing information from the spyrelet to remote monitoring processes.
    This cache is meant to be temporary, so unlike data which simply accumulates, this can be overwritten asynchronously
    """
    def reg_cache_clear(self):
        self.client['Spyre_Live_Data']['Register'].update_one({'_id':self.name},{'$set':{'cache':{}}}, upsert=True)

    def reg_cache_store(self, **kwargs):
        d = {'cache.{}'.format(key):val for key,val in custom_encode(kwargs).items()}
        if len(d):
            return self.client['Spyre_Live_Data']['Register'].update_one({'_id':self.name},{'$set':d}, upsert=True)

    def reg_cache_get(self):
        ans = self.client['Spyre_Live_Data']['Register'].find_one({'_id':self.name},{'_id':False, 'cache':True})
        return custom_decode(ans['cache']) if 'cache' in ans else {}
            
    @property
    def child_data(self):
        return self._child_data

    @property
    def data(self):
        return pd.DataFrame(self._data)

    def save(self, filename, **kwargs):
        return save_data(self, filename, **kwargs)

    def validate(self):
        # Check that the signature for main, initialize and finalize functions are the same or not defined
        init_s = inspect.signature(self.initialize)
        fina_s = inspect.signature(self.finalize)
        main_s = inspect.signature(self.main)

        for k,p in main_s.parameters.items():
            if not p.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.VAR_KEYWORD]:
                raise Exception('Invalid main signature, Spyrelets do not support positional only, keyword only or var_positional arguments')

        is_default = lambda sig: list(sig.parameters.keys()) == [CONFIG_SPYRELETS_ARGS_KEY, 'kwargs'] and \
                                 sig.parameters[CONFIG_SPYRELETS_ARGS_KEY].kind == inspect.Parameter.VAR_POSITIONAL and \
                                 sig.parameters['kwargs'].kind == inspect.Parameter.VAR_KEYWORD
        
        is_equal = lambda s1, s2: list(s1.parameters.keys()) == list(s2.parameters.keys()) and \
                                  all([s1.parameters[x].kind == s2.parameters[x].kind for x in s1.parameters.keys()])

        valid_init = is_default(init_s) or is_equal(init_s, main_s)
        valid_fina = is_default(fina_s) or is_equal(fina_s, main_s)

        if not (valid_fina and valid_init):
            raise Exception('Signature of initialize and finalize function must be the same as that of main for spyrelet: ' + self.name)

class Spyrelet_Launcher():
    def __init__(self, spyrelet):
        self.spyrelet = spyrelet
        params = spyrelet.PARAMS

        ps = inspect.signature(self.spyrelet.main).parameters
        for k, p in ps.items():
            if p.default == inspect.Parameter.empty and not k in params:
                raise Exception("Missing positional argument in PARAMS ({})".format(k))
        
        #First add all the params which require ordering
        def infer(default):
            return {'type':type(default)}
        self.params = OrderedDict([(k, params[k] if k in params else infer(p.default)) for k,p in ps.items() if not p.kind is inspect.Parameter.VAR_KEYWORD])
        for pname, pdescr in self.params.items():
            if not 'defaults' in pdescr and ps[pname].default != inspect._empty:
                self.params[pname].update(default=ps[pname].default)

        self.pos_or_kw_params = list(self.params.keys())
        #Add in all the other params
        for k, val in params.items():
            if not k in self.params:
                self.params[k] = val

        #Load defaults from client
        self.reg = self.spyrelet.client['Spyre_Live_Data']['Register']

    def get_defaults(self):
        ans = self.reg.find_one({'_id':self.spyrelet.name},{'_id':False, 'defaults':True})
        return custom_decode(ans['defaults']) if 'defaults' in ans else {}

    def run(self, progress=None, **params_dict):
        self.spyrelet.set_defaults(**params_dict)
        defaults = self.get_defaults()
        defaults.update(params_dict)
        params = {k:(val.array if type(val)==RangeDict else val) for k, val in defaults.items()}
        pos_or_kw_params = [params[k] for k in self.pos_or_kw_params]
        others = {k:val for k, val in params.items() if not k in self.pos_or_kw_params}
        return self.spyrelet.run(*pos_or_kw_params, progress=progress, **others)

def load_spyrelet_class(spyrelet_name, cfg):
    """Load a spyrelet class from a file (whose location is defined in cfg)"""
    # discover spyrelet file and class
    spyrelet_path_str,_ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY , \
                                spyrelet_name, CONFIG_SPYRELETS_FILE_KEY])
    spyrelet_class_name, spyrelet_cfg_path_str = get_config_param(cfg, \
                                [CONFIG_SPYRELETS_KEY, spyrelet_name, \
                                CONFIG_SPYRELETS_CLASS_KEY])
    # resolve the spyrelet file location
    # if the path isn't absolute resolve it relative to the config file
    spyrelet_path = Path(spyrelet_path_str)
    if not spyrelet_path.is_absolute():
        spyrelet_path = (Path(spyrelet_cfg_path_str).parent / \
                                spyrelet_path).resolve()
    return load_class_from_file(spyrelet_path, spyrelet_class_name)

def load_spyrelet(spyrelet_name, gateway, sub_spyrelet=False):
    """
    recursive function that loads a spyrelet from the config
    and also loads all sub-spyrelets
    """
    cfg = gateway.config
    if spyrelet_name in _LOADED_SPYRELETS:
        if sub_spyrelet:
            return
        else:
            raise SpyreletLoadError(None, 'spyrelet [{}] is already '
                                          'defined'.format(spyrelet_name))

    # discover any sub-spyrelets
    try:
        sub_spyrelet_names, _ = get_config_param(cfg, \
                                                 [CONFIG_SPYRELETS_KEY, spyrelet_name, \
                                                  CONFIG_SUB_SPYRELETS_KEY])
    except ConfigEntryNotFoundError:
        logging.debug('spyrelet [{}] no sub-spyrelets found'. \
                      format(spyrelet_name))
        sub_spyrelet_names = {}

    # iterate through the sub-spyrelets and load them
    sub_spyrelets = {}
    for s in sub_spyrelet_names:
        if s in _LOADED_SPYRELETS:
            sub_spyrelets[s] = _LOADED_SPYRELETS[s]
        else:
            try:
                sub_spyrelets[s] = load_spyrelet(s, gateway, sub_spyrelet=True)
            except:
                raise SpyreletLoadError(None, 'spyrelet [{}] '
                                              'sub-spyrelet [{}] failed to load'. \
                                        format(spyrelet_name, s))

    # discover and load the spyrelet class
    spyrelet_class = load_spyrelet_class(spyrelet_name, cfg)

    # discover the spyrelet devices
    try:
        dev_aliases, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, \
                                                spyrelet_name, CONFIG_DEVS_KEY])
    except ConfigEntryNotFoundError:
        logging.debug('spyrelet [{}] no device aliases found'. \
                      format(spyrelet_name))
        dev_aliases = {}

    # discover the spyrelet arguments
    try:
        args, _ = get_config_param(cfg, [CONFIG_SPYRELETS_KEY, \
                                         spyrelet_name, CONFIG_SPYRELETS_ARGS_KEY])
    except ConfigEntryNotFoundError:
        logging.debug('spyrelet [{}] no args found'.format(spyrelet_name))
        args = {}
    args = custom_decode(args)

    # create the spyrelet
    spyrelet = spyrelet_class(spyrelet_name, gateway, device_aliases=dev_aliases, spyrelets=sub_spyrelets, **args)
    _LOADED_SPYRELETS[spyrelet_name] = spyrelet
    logging.info('loaded spyrelet [{}]'.format(spyrelet_name))

    return spyrelet

def load_all_spyrelets(gateway):
    """Load all of the spyrelets from the config file"""

    # spyrelet parameters to parse
    spyrelet_configs, _ = copy(get_config_param(gateway.config, [CONFIG_SPYRELETS_KEY]))
    # check to see if any spyrelets are loaded
    if _LOADED_SPYRELETS:
        raise SpyreletLoadError(None, 'the following spyrelets were already '
                                        'loaded so nothing was done: {}'.\
                                        format(_LOADED_SPYRELETS))

    # parse the spyrelets, loading them as we go until there
    # are none left
    while bool(spyrelet_configs):
        spyrelet_name = next(iter(spyrelet_configs))
        load_spyrelet(spyrelet_name, gateway)
        # remove this spyrelet from the list of spyrelets to be loaded
        del spyrelet_configs[spyrelet_name]
    return _LOADED_SPYRELETS

def unload_spyrelet(name, client=None):
    """function to unload a current spyrelet"""
    if name not in _LOADED_SPYRELETS:
        raise SpyreletUnloadError(None, 'the spyrelet {} does not exist'.format(name))
    if client is None:
        client = get_mongo_client()
    try:
        client['Spyre_Live_Data'][name].drop()
        client['Spyre_Live_Data']['Register'].delete_one({'_id': name})
    except PyMongoError as error:
        raise SpyreletUnloadError(error, 'error attempting to remove spyrelet: {} '
                                    'from the MongoDB server'.format(name))
    _LOADED_SPYRELETS.pop(name)

def unload_all_spyrelets(except_list=[], client=None):
    """function to unload all current spyrelet(s)"""
    if client is None: client = get_mongo_client()
    all_in_reg = [x['_id'] for x in client['Spyre_Live_Data']['Register'].find({},{})]
    for name in all_in_reg:
        if not name in except_list:
            unload_spyrelet(name, client=client)

def reload_spyrelet(spyrelet_names, gateway, client=None):
    """function to reload current spyrelet(s)"""
    # if only one name is passed, convert it to list for handling
    if type(spyrelet_names) is not list: spyrelet_names = [ spyrelet_names ]

    if client is None: client = get_mongo_client()

    for name in spyrelet_names:
        unload_spyrelet(name, client=client)
        load_spyrelet(name, gateway)

def reload_all_spyrelets(gateway, except_list=[], client=None):
    """function to reload all current spyrelet(s)"""
    if client is None: client = get_mongo_client()

    # get the symmetric difference
    spyrelets = list(set(_LOADED_SPYRELETS.keys()) ^ set(except_list))

    unload_all_spyrelets(except_list, client)
    if except_list:
        for name in spyrelets:
            load_spyrelet(name, gateway)
    else:
        load_all_spyrelets(gateway)

@property
def LOADED_SPYRELETS():
    """function to return all loaded spyrelets"""
    return _LOADED_SPYRELETS



#!/usr/bin/env python
"""
CLI for setting the config files

Author: Jacob Feder
Date: 9/13/2020
"""

###########################
# imports
###########################

# std
import argparse
import logging
from pathlib import Path

# nspyre
from nspyre.definitions import CLIENT_META_CONFIG_PATH, SERVER_META_CONFIG_PATH
from nspyre.config.config_files import meta_config_add, meta_config_remove, \
                                meta_config_files, meta_config_enabled_idx, \
                                meta_config_set_enabled_idx

###########################
# globals
###########################

THIS_DIR = Path(__file__).parent
DEFAULT_LOG = THIS_DIR / 'config.log'

###########################
# exceptions
###########################

class NSpyreConfigError(Exception):
    pass

###########################
# classes / functions
###########################

def main():
    """Entry point for instrument server CLI"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-config',
                            description='Set the nspyre configuration files')
    arg_parser.add_argument('-a', '--add-config', nargs='+',
                            default=None,
                            help='permanently add a configuration '
                            'file(s) to the list to be imported on startup')
    arg_parser.add_argument('-d', '--delete-config', nargs='+',
                            default=None,
                            help='remove a configuration file(s) from '
                            'the list to be imported on startup - pass either '
                            'the file path or its index')
    arg_parser.add_argument('-l', '--list-configs',
                            action='store_true',
                            help='list the configuration files to be '
                            'imported on startup')
    arg_parser.add_argument('-q', '--quiet',
                            action='store_true',
                            help='disable logging')
    arg_parser.add_argument('-s', '--set-config',
                            default=None,
                            help='set the active config file that will be used '
                            'by nspyre')
    arg_parser.add_argument('-v', '--verbosity',
                            default='info',
                            help='the verbosity of logging - options are: '
                            'debug, info, warning, error')
    arg_parser.add_argument('client_or_inserv', help='pass [client] to modify '
        'the client configuration files, [inserv] for the instrument server')
    cmd_args = arg_parser.parse_args()

    # configure server logging behavior
    if not cmd_args.quiet:
        if cmd_args.verbosity.lower() == 'debug':
            log_level = logging.DEBUG
        elif cmd_args.verbosity.lower() == 'info':
            log_level = logging.INFO
        elif cmd_args.verbosity.lower() == 'warning':
            log_level = logging.WARNING
        elif cmd_args.verbosity.lower() == 'error':
            log_level = logging.ERROR
        else:
            raise Exception('didn\'t recognize logging level [{}]'.\
                                        format(cmd_args.verbosity)) from None

        logging.basicConfig(level=log_level,
                        format='%(asctime)s -- %(levelname)s -- %(message)s',
                        handlers=[logging.StreamHandler()])

    if cmd_args.client_or_inserv == 'client':
        meta_config_path = CLIENT_META_CONFIG_PATH
    elif cmd_args.client_or_inserv == 'inserv':
        meta_config_path = SERVER_META_CONFIG_PATH
    else:
        raise NSpyreConfigError('expected either [client] or [inserv]')

    if cmd_args.add_config:
        # the user asked us to add config files to the meta-config
        meta_config_add(meta_config_path, cmd_args.add_config)
    if cmd_args.delete_config:
        # the user asked us to remove config files from the meta-config
        meta_config_remove(meta_config_path, cmd_args.delete_config)
    if cmd_args.set_config:
        # the user asked to change the enabled config
        meta_config_set_enabled_idx(meta_config_path, cmd_args.set_config)
    if cmd_args.list_configs:
        # the user asked us to list the config files from the meta-config
        files = meta_config_files(meta_config_path)
        idx = meta_config_enabled_idx(meta_config_path)
        if files:
            for i in range(len(files)):
                print('{} {}: {}'.format('*' if i == idx else ' ', i, files[i]))
        else:
            print('no configuration files exist - use --add-config to add files')

###########################
# standalone main
###########################

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
This module handles reading and writing YAML configuration files

Author: Jacob Feder
Date: 7/25/2020
"""

###########################
# imports
###########################

# std
from pathlib import Path
from importlib import import_module
import logging

# 3rd party
import yaml

# nspyre
from nspyre.definitions import join_nspyre_path, CLIENT_META_CONFIG_PATH

###########################
# globals
###########################

META_CONFIG_FILES_ENTRY = 'config_files'
META_CONFIG_ENABLED_IDX = 'enabled'

###########################
# exceptions
###########################

class ConfigEntryNotFoundError(Exception):
    """Exception for when a configuration file doesn't contain the desired
    entry"""
    def __init__(self, config_path, msg=None):
        if msg is None:
            msg = 'Config file was expected to contain parameter: [{}] ' \
                    'but it wasn\'t found'.format(' -> '.join(config_path))
        super().__init__(msg)
        self.config_path = config_path

class ConfigError(Exception):
    """General Config file exception"""
    def __init__(self, error, msg):
        super().__init__(msg)
        if error:
            logging.exception(error)

###########################
# classes / functions
###########################

# A meta-config.yaml file contains a single entry with key 
# META_CONFIG_FILES_ENTRY and value = a list of all the config files that
# should be read

def load_raw_config(filepath):
    """Return a config file dictionary loaded from a YAML file"""
    with open(filepath, 'r') as f:
        conf = yaml.safe_load(f)
    return conf

def meta_config_add(meta_config_file, files):
    """Add config files to the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    new_files = []
    for f in files:
        f_path = Path(f).resolve()
        if not f_path.is_file():
            raise FileNotFoundError('file [{}] not found'.format(f_path))
        if str(f_path) in config_list:
            raise ConfigError(None, 'the config file {} is already available'.format(f))
        new_files.append(str(f_path))
    meta_config[META_CONFIG_FILES_ENTRY] = config_list + new_files
    write_config(meta_config, meta_config_file)

def meta_config_remove(meta_config_file, files):
    """Remove config files from the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    enabled_idx = meta_config[META_CONFIG_ENABLED_IDX]
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    # list of indicies to remove from config_list
    pop_list = []
    # go through the list of config files / indicies and generate
    # a list of indicies to remove
    for c in files:
        try:
            idx = int(c)
        except ValueError:
            # the user passed the config name as a string, so we should first
            # find its index
            try:
                idx = meta_config[META_CONFIG_FILES_ENTRY].index(c)
            except ValueError as exc:
                raise ConfigError(exc, 'config file [{}] was not found in the '
                                'meta-config - check that the file path shown '
                                'using --list-configs matches given input'.\
                                format(c)) from None
        pop_list.append(idx)

    pop_list.sort(reverse=True)
    new_enabled_idx = enabled_idx
    for idx in pop_list:
        try:
            config_list.pop(idx)
            # if an item before the enabled one was deleted we have to decrement
            # the enabled index
            if idx < enabled_idx:
                new_enabled_idx -= 1
        except IndexError as exc:
            raise ConfigError(exc, 'tried to remove config file index [{}] '
                'that was out of range'.format(idx))

    # if the user removed the currently enabled config
    if enabled_idx in pop_list:
        new_enabled_idx = 0

    meta_config[META_CONFIG_FILES_ENTRY] = config_list
    meta_config[META_CONFIG_ENABLED_IDX] = new_enabled_idx
    write_config(meta_config, meta_config_file)

def meta_config_files(meta_config_file):
    """Return the paths of the config files in the meta-config"""
    meta_config = load_raw_config(meta_config_file)
    config_list = meta_config[META_CONFIG_FILES_ENTRY]
    return config_list

def meta_config_enabled_idx(meta_config_file):
    """Return the index of the enabled config"""
    meta_config = load_raw_config(meta_config_file)
    return meta_config[META_CONFIG_ENABLED_IDX]

def meta_config_set_enabled_idx(meta_config_file, idx_or_str):
    """Return the index of the enabled config"""
    meta_config = load_raw_config(meta_config_file)
    try:
        idx = int(idx_or_str)
    except ValueError:
        # the user passed the config name as a string, so we should first
        # find its index
        try:
            idx = meta_config[META_CONFIG_FILES_ENTRY].index(idx_or_str)
        except ValueError as exc:
            raise ConfigError(exc, 'config file [{}] was not found in the '
                'meta-config - check that the file path shown '
                'using --list-configs matches given input'.\
                format(idx_or_str)) from None
    meta_config[META_CONFIG_ENABLED_IDX] = idx
    write_config(meta_config, meta_config_file)

def load_meta_config(meta_config_path=None):
    """Takes a 'meta' config file and returns the file path of the activated 
    config file"""

    # TODO this logic should be removed and meta_config_path always passed in
    if not meta_config_path:
        meta_config_path = CLIENT_META_CONFIG_PATH

    # load the meta config
    meta_config = load_raw_config(meta_config_path)
    # get the index of the enabled config file
    enabled_idx = meta_config[META_CONFIG_ENABLED_IDX]

    # get the config file path
    config_files = meta_config[META_CONFIG_FILES_ENTRY]
    if not config_files:
        raise ConfigError(None, 'no configuration files exist - '
                    'use nspyre-config --add-config to add files') from None
    cfg_path = Path(config_files[enabled_idx])

    # resolve relative paths
    if not cfg_path.is_absolute():
        cfg_path = meta_config_path.parent / cfg_path

    return cfg_path

def load_config(cfg_path):
    """Load a config file and return the configuration as a dictionary"""
    try:
        cfg_dict = {str(cfg_path): load_raw_config(cfg_path)}
    except FileNotFoundError as exc:
        raise ConfigError(exc, 'configuration file [{}] doesn\'t exist'.\
                            format(cfg_path)) from None
    return cfg_dict

def write_config(config_dict, filepath):
    """Write a dictionary to a YAML file"""
    # open the file and write it's config dictionary
    with open(filepath, 'w') as file:
        yaml.dump(config_dict, file, default_flow_style=False)

def get_config_param(config_dict, path):
    """Navigate a YAML-loaded config file and return a particular parameter 
    given by 'path'. If multiple config files contain the first element of
    'path', this will attempt to navigate the the first config file it finds
    that contains the first element of 'path'."""
    first_elem = path[0]
    for conf in config_dict:
        # first find the config file containing the first path element
        loc = config_dict[conf]
        if first_elem in loc:
            # now descend into the config dictionary, following the keys
            # one-by-one in path
            for p in path:
                try:
                    loc = loc[p]
                except KeyError:
                    raise ConfigEntryNotFoundError(path) from None
            return loc, conf
    # if we reach this point the first path element wasn't found in any
    # config file entries
    raise ConfigEntryNotFoundError(path) from None



#!/usr/bin/env python
"""
This module serves a shell prompt allowing the user runtime control of 
the instrument server

Author: Jacob Feder
Date: 7/8/2020
"""

###########################
# imports
###########################

# std
import argparse
from cmd import Cmd
import logging
import pdb
from pathlib import Path

# 3rd party
import pyvisa

# nspyre
from nspyre.inserv.inserv import InstrumentServer
from nspyre.config.config_files import load_meta_config
from nspyre.definitions import SERVER_META_CONFIG_PATH

###########################
# globals
###########################

THIS_DIR = Path(__file__).parent
DEFAULT_LOG = THIS_DIR / 'inserv.log'

###########################
# classes / functions
###########################

class InservCmdPrompt(Cmd):
    """Instrument Server shell prompt processor"""
    def __init__(self, inserv):
        super().__init__()
        self.inserv = inserv

    def emptyline(self):
        """When no command is entered"""
        pass

    def do_list(self, arg_string):
        """List all the available devices"""
        if arg_string:
            print('Expected 0 args')
            return
        for d in self.inserv._devs.keys():
            print(d)

    def do_config(self, arg_string):
        """Reload the server config files"""
        if arg_string:
            print('Expected 0 args')
            return
        # attempt to reload the config files
        try:
            self.inserv.update_config(config_file=\
                                    args[0] if arg_string else None)
        except Exception as exc:
            logging.exception(exc)
            print('Failed to reload config files')
            return

    def do_dev(self, arg_string):
        """Restart the connection with a device\n<string> the device name"""
        args = arg_string.split(' ')
        if not arg_string or len(args) > 1:
            print('Expected 1 arg: device name')
            return
        dev_name = args[0]
        try:
            self.inserv.reload_device(dev_name)
        except Exception as exc:
            logging.exception(exc)
            print('Failed to reload device [{}]'.format(dev_name))
            return

    def do_dev_all(self, arg_string):
        """Restart the connection with all devices"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.reload_devices()
        except Exception as exc:
            logging.exception(exc)
            print('Failed to reload all devices')
            return

    def do_restart(self, arg_string):
        """Restart the server AND reload the config file and all devices"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.restart()
        except Exception as exc:
            logging.exception(exc)
            print('Failed to restart')
            return

    def do_server_restart(self, arg_string):
        """Restart the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.reload_server()
        except Exception as exc:
            logging.exception(exc)
            print('Failed to restart server')
            return

    def do_server_stop(self, arg_string):
        """Stop the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.stop_server()
        except Exception as exc:
            logging.exception(exc)
            print('Failed to stop server')
            return

    def do_server_start(self, arg_string):
        """Start the rpyc server"""
        if arg_string:
            print('Expected 0 args')
            return
        try:
            self.inserv.start_server()
        except Exception as exc:
            logging.exception(exc)
            print('Failed to start server')
            return

    def do_debug(self, arg_string):
        """Drop into the debugging console"""
        if arg_string:
            print('Expected 0 args')
            return
        pdb.set_trace()
        
    def do_quit(self, arg_string):
        """Quit the program"""
        if arg_string:
            print('Expected 0 args')
            return
        logging.info('exiting...')
        # close all open resources
        self.inserv.stop_server()
        for dev_name in list(self.inserv._devs):
            self.inserv.del_device(dev_name)
        pyvisa.ResourceManager().close()
        raise SystemExit

def main():
    """Entry point for instrument server CLI"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-inserv',
                            description='Run an nspyre instrument server')
    arg_parser.add_argument('-l', '--log',
                            default=DEFAULT_LOG,
                            help='log to the provided file location')
    arg_parser.add_argument('-m', '--mongo',
                            default=None,
                            help='use the provided mongodb address rather than '
                            'the one listed in the config (e.g. '
                            'mongodb://192.168.1.27:27017/)') 
    arg_parser.add_argument('-q', '--quiet',
                            action='store_true',
                            help='disable logging')
    arg_parser.add_argument('-v', '--verbosity',
                            default='debug',
                            help='the verbosity of logging - options are: '
                            'debug, info, warning, error')
    cmd_args = arg_parser.parse_args()

    # configure server logging behavior
    if not cmd_args.quiet:
        if cmd_args.verbosity.lower() == 'debug':
            log_level = logging.DEBUG
        elif cmd_args.verbosity.lower() == 'info':
            log_level = logging.INFO
        elif cmd_args.verbosity.lower() == 'warning':
            log_level = logging.WARNING
        elif cmd_args.verbosity.lower() == 'error':
            log_level = logging.ERROR
        else:
            raise InstrumentServerError('didn\'t recognize logging level [{}]'.\
                                        format(cmd_args.verbosity)) from None

        logging.basicConfig(level=log_level,
                        format='%(asctime)s -- %(levelname)s -- %(message)s',
                        handlers=[logging.FileHandler(cmd_args.log, 'w+'),
                                logging.StreamHandler()])

    # init and start RPyC server
    logging.info('starting instrument server...')
    config_path = load_meta_config(SERVER_META_CONFIG_PATH)
    inserv = InstrumentServer(config_path, cmd_args.mongo)

    # start the shell prompt event loop
    cmd_prompt = InservCmdPrompt(inserv)
    cmd_prompt.prompt = 'inserv > '
    cmd_prompt.cmdloop('instrument server started...')

###########################
# standalone main
###########################

if __name__ == '__main__':
    main()


from nspyre.inserv.gateway import InservGateway
import pdb

if __name__ ==  '__main__':
    ig = InservGateway()
    sg = ig.local1._devs['fake_tcpip_sg']
    sg.offset = 1
    print(sg.offset)
    #pdb.set_trace()



"""
This module manages and centralizes connections to one or more instrument
servers. All instrument server connections should be done through the
InservGateway class.

Author: Jacob Feder
Date: 7/11/2020
"""

###########################
# imports
###########################

# std
import os
import logging

# 3rd party
import parse
import pymongo
from pymongo.errors import PyMongoError
import rpyc

# nspyre
from nspyre.utils.misc import register_quantity_brining
from nspyre.config.config_files import get_config_param, load_config
from nspyre.definitions import MONGO_CONNECT_TIMEOUT, \
                            MONGO_SERVERS_KEY, MONGO_SERVERS_SETTINGS_KEY, \
                            MONGO_RS, RPYC_CONN_TIMEOUT, RPYC_SYNC_TIMEOUT, \
                            INSERV_DEV_ACCESSOR
from nspyre.definitions import Q_

# for properly serializing/deserializing quantity objects using the local
# pint unit registry
register_quantity_brining(Q_)

###########################
# exceptions
###########################

class InservGatewayError(Exception):
    """General InservGateway exception"""
    def __init__(self, msg):
        super().__init__(msg)

###########################
# classes / functions
###########################

# Temporary monkey patching of rpyc to implement synchronous about_to_disconnect feature
# Need to define consts.HANDLE_ABOUT_TO_CLOSE before consts is import by protocol (this is done
# in the load of nspyre.inser.inserv)
import nspyre.inserv.inserv
from rpyc.core.protocol import consts

# Need to monkey patch VoidService for rpyc.utils.factory.connect_stream
from rpyc.core.service import Service
Service._protocol = nspyre.inserv.inserv.InstrumentConnection
Service = nspyre.inserv.inserv.InstrumentService
from rpyc.core.service import VoidService
VoidService = nspyre.inserv.inserv.VoidInstrumentService


class InservGateway():
    """Loads a configuration file, then attempts to connect to all 
    instrument servers

    This class is a representation of the instrument server from the
    client's perspective. It contains server connection information, and an
    instance variable object for each device connected to the remote instrument
    server
    """
    def __init__(self, config_file, mongo_addr=None):
        # config dictionary
        self.config = {}
        # dictionary of available rpyc instrument servers
        # key is the server string id, value is a tuple (rpyc conn, bg thread)
        # e.g. {'local1': (rpyc.core.protocol.Connection, 
        #                  rpyc.utils.helpers.BgServingThread),
        #       'remote1': ...}
        self._servers = {}
        self.mongo_addr = None
        self.mongo_client = None
        self.config = None
        self.update_config(config_file)
        self.config_mongo(mongo_addr)
        self.connect_servers()

    def config_mongo(self, mongo_addr=None):
        """Set up the mongodb database (but connection won't take place until
        a query is made)"""
        if mongo_addr:
            self.mongo_addr = mongo_addr
        else:
            self.mongo_addr,_ = get_config_param(self.config, ['mongodb_addr'])
        logging.info('connecting to mongodb server [{}]...'.\
                            format(self.mongo_addr))
        self.mongo_client = pymongo.MongoClient(mongo_addr,
                            replicaset=MONGO_RS,
                            serverSelectionTimeoutMS=MONGO_CONNECT_TIMEOUT)

    def connect_servers(self):
        """Auto discover and attempt connection to all of the instrument
        servers in mongodb"""
        # retrieve all of the instrument server settings from mongo
        try:
            all_db_names = self.mongo_client.list_database_names()
        except PyMongoError:
            raise InservGatewayError('Failed connecting to mongodb [{}]'.\
                                        format(self.mongo_addr)) from None
        logging.info('connected to mongodb server [{}]'.format(self.mongo_addr))
        for db_name in all_db_names:
            server_name = parse.parse(MONGO_SERVERS_KEY, db_name)
            # filter out all of the dbs that aren't instrument servers
            if server_name:
                # if the name was extracted successfully,
                # get it's value as a string
                server_name = server_name[0]
                # retrieve the server settings dictionary
                db_entry = self.mongo_client[db_name]\
                                    [MONGO_SERVERS_SETTINGS_KEY].find_one()
                self.connect_server(server_name,
                                    db_entry['address'],
                                    db_entry['port'])

    def disconnect_servers(self):
        """Attempt disconnection from all of the instrument servers"""
        for s in list(self._servers):
            self.disconnect_server(s)

    def connect_server(self, s_id, s_addr, s_port):
        """Attempt connection to an instrument server"""
        try:
            # connect to the rpyc server running on the instrument server
            # and start up a background thread to fullfill requests on the
            # client side
            conn = rpyc.connect(s_addr, s_port,
                            config={'allow_pickle' : True,
                                    'timeout' : RPYC_CONN_TIMEOUT,
                                    'sync_request_timeout': RPYC_SYNC_TIMEOUT})

            bg_serving_thread = rpyc.BgServingThread(conn)
            
            # this allows the instrument server to have full access to this
            # client's object dictionaries - appears necessary for lantz
            conn._config['allow_all_attrs'] = True

            self._servers[s_id] = (conn, bg_serving_thread)
        except BaseException:
            raise InservGatewayError('Failed to connect to '
                            'instrument server [{}] at address [{}]'.\
                            format(s_id, s_addr)) from None
        logging.info('instrument server gateway connected to instrument '
                    'server [{}]'.format(s_id))

    def disconnect_server(self, s_id):
        """Disconnect from an instrument server and remove it's associated 
        devices"""
        try:
            conn = self._servers[s_id][0]
            bg_serving_thread = self._servers[s_id][1]
            bg_serving_thread.stop()
            conn.close()
            del self._servers[s_id]
        except BaseException:
            raise InservGatewayError('Failed to disconnect from '
                            'instrument server [{}]'.format(s_id)) from None
        logging.info('instrument server gateway disconnected '
                        'from server [{}]'.format(s_id))

    def servers(self):
        """Return a dictionary containing 'server name' mapped to
        an rpyc conn object"""
        servers_dict = {}
        for s in self._servers:
            servers_dict[s] = self._servers[s][0]

        return servers_dict

    def __getattr__(self, attr):
        """Allow the user to access the server objects directly using
        e.g. gateway.server1.sig_gen.frequency notation"""
        if attr in self._servers:
            return self._servers[attr][0].root
        else:
            raise AttributeError('\'{}\' object has no attribute \'{}\''.\
                        format(self.__class__.__name__, attr))

    def update_config(self, filename):
        """Reload the config file"""
        self.config = load_config(filename)

    def __enter__(self):
        """Python context manager setup"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Python context manager teardown"""
        self.disconnect_servers()

if __name__ == '__main__':
    # configure server logging behavior
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s -- %(levelname)s -- %(message)s',
                        handlers=[logging.StreamHandler()])
    # TODO unit testing module
    with InservGateway() as im:
        pass
        # sg_loc = 'local1/fake_sg'
        # im.devs[sg_loc].amplitude = Q_(2.0, 'volt')
        # im.devs[sg_loc].amplitude = Q_(10.0, 'volt')
        # im.devs[sg_loc].dout[1] = Q_(0.0, 'volt')
        # im.devs[sg_loc].dout[2] = Q_(2.0, 'volt')
        # im.devs[sg_loc].dout[3] = Q_(3.0, 'volt')
        # im.devs[sg_loc].dout[1] = Q_(1.0, 'volt')
        #print('found devices:\n{}'.format(im.devs))
        #print(Q_(5, 'volt') + im.devs[sg_loc].amplitude)



"""
This module:
    - loads the server config file
    - connects to all of the instruments specified in the config files
    - creates a RPyC (python remote procedure call) server to allow remote
        machines to access the instruments (should be done using InservGateway)

Author: Jacob Feder
Date: 7/8/2020
"""

###########################
# imports
###########################

# std
from pathlib import Path
import logging
import threading
import time
from functools import partial
import copy
import functools
import weakref

# 3rd party
import rpyc
from rpyc.utils.server import ThreadedServer
import pymongo
from lantz import Q_, DictFeat
from pimpmyclass.helpers import DictPropertyNameKey
from pint import Quantity

# nspyre
from nspyre.utils.misc import register_quantity_brining
from nspyre.config.config_files import get_config_param, load_config, \
                                    meta_config_add, meta_config_remove, \
                                    meta_config_files, ConfigEntryNotFoundError
from nspyre.utils.misc import load_class_from_str, load_class_from_file
from nspyre.definitions import MONGO_SERVERS_KEY, \
                MONGO_SERVERS_SETTINGS_KEY, MONGO_CONNECT_TIMEOUT, \
                MONGO_RS, RPYC_SYNC_TIMEOUT, CONFIG_MONGO_ADDR_KEY, \
                join_nspyre_path

# for properly serializing/deserializing quantity objects using the local
# pint unit registry
register_quantity_brining(Q_)

###########################
# globals
###########################

CONFIG_SERVER_SETTINGS = 'server_settings'
CONFIG_SERVER_DEVICES = 'devices'
CONFIG_SERVER_DEVICE_LANTZ_CLASS = 'lantz_class'
CONFIG_SERVER_DEVICE_CLASS_FILE = 'class_file'
CONFIG_SERVER_DEVICE_CLASS_NAME = 'class'

RPYC_SERVER_STOP_EVENT = threading.Event()

###########################
# exceptions
###########################

class InstrumentServerError(Exception):
    """General InstrumentServer exception"""
    def __init__(self, error, msg):
        super().__init__(msg)
        if error:
            logging.exception(error)

###########################
# classes
###########################


rpyc.core.consts.HANDLE_ABOUT_TO_CLOSE = 21
from rpyc.core.protocol import consts


class InstrumentConnection(rpyc.Connection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _pre_cleanup(self):  # IO
        self._local_root.on_about_to_disconnect(self)

    def close(self, _catchall=True):  # IO
        """closes the connection, releasing all held resources"""
        if self._closed:
            return
        self._closed = True
        try:
            self.sync_request(consts.HANDLE_ABOUT_TO_CLOSE)
            self._async_request(consts.HANDLE_CLOSE)
        except EOFError:
            pass
        except Exception:
            if not _catchall:
                raise
        finally:
            self._cleanup(_anyway=True)

    @classmethod
    def _request_handlers(cls):  # request handlers
        return {
            consts.HANDLE_PING: cls._handle_ping,
            consts.HANDLE_CLOSE: cls._handle_close,
            consts.HANDLE_GETROOT: cls._handle_getroot,
            consts.HANDLE_GETATTR: cls._handle_getattr,
            consts.HANDLE_DELATTR: cls._handle_delattr,
            consts.HANDLE_SETATTR: cls._handle_setattr,
            consts.HANDLE_CALL: cls._handle_call,
            consts.HANDLE_CALLATTR: cls._handle_callattr,
            consts.HANDLE_REPR: cls._handle_repr,
            consts.HANDLE_STR: cls._handle_str,
            consts.HANDLE_CMP: cls._handle_cmp,
            consts.HANDLE_HASH: cls._handle_hash,
            consts.HANDLE_INSTANCECHECK: cls._handle_instancecheck,
            consts.HANDLE_DIR: cls._handle_dir,
            consts.HANDLE_PICKLE: cls._handle_pickle,
            consts.HANDLE_DEL: cls._handle_del,
            consts.HANDLE_INSPECT: cls._handle_inspect,
            consts.HANDLE_BUFFITER: cls._handle_buffiter,
            consts.HANDLE_OLDSLICING: cls._handle_oldslicing,
            consts.HANDLE_CTXEXIT: cls._handle_ctxexit,
            consts.HANDLE_ABOUT_TO_CLOSE: cls._handle_about_to_close,
        }

    def _handle_about_to_close(self):  # request handler
        self._pre_cleanup()


class InstrumentService(rpyc.Service):

    _protocol = InstrumentConnection

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def on_about_to_disconnect(self, conn):
        """called when the connection had already terminated for cleanup
        (must not perform any IO on the connection)"""
        pass


class VoidInstrumentService(InstrumentService):
    """void service - an do-nothing service"""
    __slots__ = ()


rpyc.Connection = InstrumentConnection
rpyc.Service = InstrumentService
rpyc.VoidService = VoidInstrumentService


class InstrumentServer(InstrumentService):
    """RPyC provider that loads lantz devices and exposes them to the remote
    client
    """

    def __init__(self, config_file, mongo_addr=None):
        super().__init__()
        # lantz devices
        self._devs = {}
        # configuration
        self.config = {}
        self.config_file = None
        self.name = None
        # server settings
        self.ip = None
        self.port = None
        # rpyc server object
        self._rpyc_server = None
        # mongodb
        self.db_name = None
        self.mongo_addr = None
        self.mongo_client = None
        self.db = None

        # storage container for hook methods that update mongodb feats when
        # lantz feats are changed - this is required because pysignal uses
        # weakrefs, so any local functions you pass to it will fail when they
        # go out of scope
        # TODO replace with partials
        self.feat_hook_functions = {}
        self.dictfeat_hook_functions = {}

        self.update_config(config_file)
        self.reload_server_config()
        self.start_server()
        self.reload_devices()


    def __getattr__(self, name):
        """Allow the user to access the driver objects directly using
        e.g. inserv.fake_tcpip_sg.amplitude notation
        """
        if name in self._devs:
            return self._devs[name]
        else:
            return self.__getattribute__(name)
            #raise AttributeError('\'{}\' object has no attribute \'{}\''.\
            #            format(self.__class__.__name__, name))


    def on_connect(self, conn):
        """Called when a client connects to the RPyC server"""
        logging.info('client [{}] connected'.format(conn))

    def on_about_to_disconnect(self, conn):
        logging.info('client [{}] about to disconnect'.format(conn))
        for device_name, device in self._devs.items():
            # iterate over all feats and dictfeats
            for attr_name, attr in list(device._lantz_feats.items()) + list(device._lantz_dictfeats.items()):
                if isinstance(attr_name, DictPropertyNameKey):
                    # filter out weird dictfeat feats
                    continue
                # access the PySignal slots
                attr_slots = getattr(device, attr_name + '_changed')._slots
                for slot in attr_slots:
                    if isinstance(slot, functools.partial) and slot.__name__ == 'InstrumentManager_getattr_func':
                        getattr(device, attr_name + '_changed').disconnect(slot)


    def on_disconnect(self, conn):
        """Called when a client disconnects from the RPyC server"""
        logging.info('client [{}] disconnected'.format(conn))

        # when an RPyC client disconnects, there are dangling netrefs
        # left over in PySignal _slots that point to objects on the client
        # side which are now inaccessible
        # this block detects and removes the references by attempting to
        # access them with a try/except
        # TODO this logic should really be implemented in PySignal during emit()
        for device_name, device in self._devs.items():
            # iterate over all feats and dictfeats
            for attr_name, attr in list(device._lantz_feats.items()) + list(device._lantz_dictfeats.items()):
                if isinstance(attr_name, DictPropertyNameKey):
                    # filter out weird dictfeat feats
                    continue
                # access the PySignal slots
                attr_slots = getattr(device, attr_name + '_changed')._slots
                for index, slot in enumerate(attr_slots):
                    try:
                        # trying to compare slot to something will force RPyC
                        # to attempt to retrieve it
                        if slot == None:
                            pass
                    except EOFError:
                        del attr_slots[index]


    def restart(self, config_file=None, mongo_addr=None):
        """Restart the server AND reload the config file and all devices"""
        logging.info('restarting...')
        self.update_config(config_file)
        self.reload_server_config()
        self.reload_devices()
        self.reload_server()


    def reload_server_config(self):
        """Reload RPyC server settings from the config"""
        self.name, _ = get_config_param(self.config, \
                        [CONFIG_SERVER_SETTINGS, 'name'])
        self.ip, _ = get_config_param(self.config, \
                        [CONFIG_SERVER_SETTINGS, 'ip'])
        self.port, _ = get_config_param(self.config, \
                        [CONFIG_SERVER_SETTINGS, 'port'])


    def connect_mongo(self, mongo_addr=None):
        """Config and connect to the mongodb database"""
        self.db_name = MONGO_SERVERS_KEY.format(self.name)
        if mongo_addr:
            self.mongo_addr = mongo_addr
        else:
            self.mongo_addr,_ = get_config_param(self.config, \
                                                [CONFIG_MONGO_ADDR_KEY])
        logging.info('connecting to mongodb server [{}]...'.\
                        format(self.mongo_addr))
        self.mongo_client = pymongo.MongoClient(mongo_addr,
                            replicaset=MONGO_RS,
                            serverSelectionTimeoutMS=MONGO_CONNECT_TIMEOUT)
        self.db = self.mongo_client[self.db_name]
        # mongodb doesn't do any actual database queries until you try to
        # perform an action on the database, so this is the point where
        # connection will occur
        try:
            current_db = self.mongo_client[self.db_name]\
                                    [MONGO_SERVERS_SETTINGS_KEY].find_one()
        except Exception as exc:
            raise InstrumentServerError(exc, 'Failed connecting to mongodb '
                                    '[{}]'.format(self.mongo_addr)) from None
        try:
            current_db_address = current_db['address']
        except:
            current_db_address = self.ip
        # check that there isn't already an instrument server with same
        # name but different address
        if current_db_address != self.ip:
            raise InstrumentServerError('An instrument server with the name '
                    '[{}] and a different address [{}] is already present on '
                    'mongodb. Instrument servers must have unique names.'.\
                    format(self.name, current_db_address)) from None
        self.mongo_client.drop_database(self.db_name)
        # add a special settings document so clients can auto-detect us
        self.db[MONGO_SERVERS_SETTINGS_KEY].insert_one({'address' : self.ip,
                                                    'port' : self.port})
        logging.info('connected to mongodb server [{}]'.format(self.mongo_addr))


    def disconnect_mongo(self):
        """Disconnect from the mongodb database"""
        # remove the database entry from mongo
        self.mongo_client.drop_database(self.db_name)
        # disconnect
        self.mongo_client.close()


    def add_device(self, dev_name):
        """Add and initialize a device"""

        # load the device parameters from the config file

        # first try getting the driver's lantz class
        try:
            dev_lantz_class, dev_class_cfg_file = get_config_param(self.config,
                                        [CONFIG_SERVER_DEVICES, dev_name, \
                                        CONFIG_SERVER_DEVICE_LANTZ_CLASS])
            try:
                dev_class = load_class_from_str('lantz.drivers.' + \
                                                dev_lantz_class)
            except Exception as exc:
                raise InstrumentServerError(exc, 'The specified lantz driver '
                    '[{}] for device [{}] couldn\'t be loaded'.\
                    format(dev_lantz_class, dev_name))
        except ConfigEntryNotFoundError:
            # if the lantz class isn't defined, try getting an absolute file
            # path and class name
            try:
                dev_class_file_str, dev_class_file_cfg_file = \
                                    get_config_param(self.config,
                                            [CONFIG_SERVER_DEVICES, dev_name,
                                            CONFIG_SERVER_DEVICE_CLASS_FILE])
                dev_class_name, _ = get_config_param(self.config,
                                            [CONFIG_SERVER_DEVICES, dev_name,
                                            CONFIG_SERVER_DEVICE_CLASS_NAME])
            except ConfigEntryNotFoundError as exc:
                raise InstrumentServerError(exc, 'The device [{}] didn\'t '
                    'contain an entry for either a lantz class "{}" or a file '
                    'path "{}" / class name "{}" to define it\'s driver type'.\
                    format(dev_name, CONFIG_SERVER_DEVICE_LANTZ_CLASS,
                            CONFIG_SERVER_DEVICE_CLASS_FILE,
                            CONFIG_SERVER_DEVICE_CLASS_NAME))
            dev_class_path = Path(dev_class_file_str)
            # resolve relative paths
            if not dev_class_path.is_absolute():
                dev_class_path = (Path(dev_class_file_cfg_file).parent / \
                                    dev_class_path).resolve()
            try:
                dev_class = load_class_from_file(dev_class_path, dev_class_name)
            except Exception as exc:
                raise InstrumentServerError(exc, 'The specified class [{}] '
                    'from file [{}] for device [{}] couldn\'t be loaded'.\
                    format(dev_class_name, dev_class_path, dev_name))

        dev_args, _ = get_config_param(self.config,
                                    [CONFIG_SERVER_DEVICES, dev_name, 'args'])
        dev_kwargs, _ = get_config_param(self.config,
                                    [CONFIG_SERVER_DEVICES, dev_name, 'kwargs'])

        # get an instance of the device
        try:
            self._devs[dev_name] = dev_class(*dev_args, **dev_kwargs)
        except Exception as exc:
            raise InstrumentServerError(exc, 'Failed to get instance of device '
                                        '{} of class {}'.\
                                        format(dev_name, dev_class)) from None

        # collect all of the lantz feature attributes
        feat_attr_list = []
        for feat_name, feat in list(dev_class._lantz_feats.items()) + list(dev_class._lantz_dictfeats.items()):
            if isinstance(feat_name, DictPropertyNameKey):
                # filter out weird dictfeat feats
                continue

            attrs = copy.deepcopy(feat.__dict__['_config'])
            if isinstance(feat, DictFeat) and 'keys' in feat.__dict__:
                attrs['keys'] = feat.__dict__['keys']

            if attrs['limits']:
                if len(attrs['limits']) == 1:
                    limits = [0, attrs['limits'][0]]
                else:
                    limits = attrs['limits']
            else:
                limits = None

            if attrs['values']:
                values = list(attrs['values'])
            else:
                values = None

            if 'keys' in attrs and attrs['keys']:
                keys = list(attrs['keys'])
            else:
                keys = None

            if 'keys' in attrs and attrs['keys'] and isinstance(feat, DictFeat):
                value = [None] * len(attrs['keys'])
            else:
                value = None

            feat_attr_list.append({
                'name':     feat_name,
                'type':     'dictfeat' if isinstance(feat, DictFeat) \
                                            else 'feat',
                'readonly': feat.__dict__['fset'] is None,
                'units':    attrs['units'],
                'limits':   limits,
                'values':   values,
                'keys':     keys,
                'value':    value
            })

            # add a custom hook for updating mongodb whenever the feat/dictfeat is written
            if isinstance(feat, DictFeat):
                def update_mongo_dictfeat(value, old_value, key, attr=feat_name, keys=keys):
                    logging.debug('{}[{}]: {} -> {}'.format(attr, key, old_value, value))
                    if isinstance(value, Quantity):
                        value = value.to(self._devs[dev_name]._lantz_dictfeats[attr]._kwargs['units']).m
                    self.db[dev_name].update_one({'name': attr},
                                                 {'$set': {'value.{}'.format(keys.index(key)): value}},
                                                 upsert=True)

                self.dictfeat_hook_functions[feat_name] = update_mongo_dictfeat
                getattr(self._devs[dev_name], feat_name + '_changed').connect(self.dictfeat_hook_functions[feat_name])
            else:
                def update_mongo_feat(value, old_value, attr=feat_name):
                    logging.debug('{}: {} -> {}'.format(attr, old_value, value))
                    if isinstance(value, Quantity):
                        value = value.to(self._devs[dev_name]._lantz_feats[attr]._kwargs['units']).m
                    self.db[dev_name].update_one({'name': attr},
                                                 {'$set': {'value': value}},
                                                 upsert=True)

                self.feat_hook_functions[feat_name] = update_mongo_feat
                getattr(self._devs[dev_name], feat_name + '_changed').connect(self.feat_hook_functions[feat_name])

        for action_name, action in dev_class._lantz_actions.items():
            feat_attr_list.append({'name' : action_name, 'type' : 'action'})

        self.db[dev_name].drop()
        # add all of the lantz feature attributes to the database
        self.db[dev_name].insert_many(feat_attr_list)

        # initialize the device
        try:
            self._devs[dev_name].initialize()
        except Exception as exc:
            logging.error(exc)
            self._devs.pop(dev_name)
            logging.error('device [{}] initialization sequence failed'.\
                            format(dev_name))
            return

        logging.info('added device [{}] with args: {} kwargs: {}'.\
                        format(dev_name, dev_args, dev_kwargs))


    def del_device(self, dev_name):
        """Remove and finalize a device"""
        try:
            self._devs.pop(dev_name).finalize()
        except Exception as exc:
            raise InstrumentServerError(exc, 'Failed deleting device [{}]'.\
                                        format(dev_name)) from None
        logging.info('deleted [{}]'.format(dev_name))


    def reload_device(self, dev_name):
        """Remove a device, then reload it from the stored config"""
        if dev_name in self._devs:
            self.del_device(dev_name)
        self.add_device(dev_name)


    def reload_devices(self):
        """Reload all devices"""
        devs, _ = get_config_param(self.config, ['devices'])
        for dev_name in devs:
            self.reload_device(dev_name)
        logging.info('reloaded all devices')


    def update_config(self, config_file=None):
        """Reload the config files"""
        # update the config with a new file if one was passed as argument
        if config_file:
            self.config_file = config_file
        # reload the config dictionary
        self.config = load_config(self.config_file)
        logging.info('loaded config files {}'.\
                        format(list(self.config.keys())))


    def reload_server(self):
        """Restart the RPyC server"""
        self.stop_server()
        self.start_server()


    def start_server(self):
        """Start the RPyC server"""
        if self._rpyc_server:
            logging.warning('can\'t start the rpyc server because one '
                            'is already running')
            return
        self.connect_mongo()
        thread = threading.Thread(target=self._rpyc_server_thread)
        thread.start()
        # wait for the server to start
        while not (self._rpyc_server and self._rpyc_server.active):
            time.sleep(0.1)


    def stop_server(self):
        """Stop the RPyC server"""
        self.disconnect_mongo()
        if not self._rpyc_server:
            logging.warning('can\'t stop the rpyc server because there '
                            'isn\'t one running')
            return
        logging.info('stopping RPyC server...')
        self._rpyc_server.close()
        # wait for the server to stop
        while self._rpyc_server.active:
            time.sleep(0.1)
        RPYC_SERVER_STOP_EVENT.wait()
        RPYC_SERVER_STOP_EVENT.clear()
        self._rpyc_server = None


    def _rpyc_server_thread(self):
        """Thread for running the RPyC server asynchronously"""
        logging.info('starting RPyC server...')
        self._rpyc_server = ThreadedServer(self, port=self.port,
                        protocol_config={'allow_pickle' : True,
                                    'allow_all_attrs' : True,
                                    'allow_setattr' : True,
                                    'allow_delattr' : True,
                                    'sync_request_timeout' : RPYC_SYNC_TIMEOUT})
        self._rpyc_server.start()
        logging.info('RPyC server stopped')
        RPYC_SERVER_STOP_EVENT.set()



#!/usr/bin/env python
"""
    nspyre.gui.__init__.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    This module runs the main nspyre GUI, which allows the user to easily
    launch other tools

    Author: Alexandre Bourassa
    Modified: Jacob Feder 7/25/2020, Michael Solomon 9/12/2020
"""

import argparse
import logging
from pathlib import Path
import sys

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

from .app import NSpyreApp
from .main_window import NSpyreMainWindow

THIS_DIR = Path(__file__).parent
DEFAULT_LOG = THIS_DIR / 'nspyre.log'

###########################
# standalone main
###########################

def main(args=None):
    """Entry point for the application script"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre',
                            usage='%(prog)s [options]',
                            description='Run the nspyre GUI')
    arg_parser.add_argument('-c', '--config', nargs='+',
                            default=None,
                            help='permanently add a configuration '
                            'file(s) to the list to be imported on startup')
    arg_parser.add_argument('-d', '--delete-config', nargs='+',
                            default=None,
                            help='remove a configuration file(s) from '
                            'the list to be imported on startup')
    arg_parser.add_argument('-e', '--list_configs',
                            action='store_true',
                            help='list the configuration files to be '
                            'imported on startup')
    arg_parser.add_argument('-l', '--log',
                            default=DEFAULT_LOG,
                            help='log to the provided file location')
    arg_parser.add_argument('-m', '-- ',
                            default=None,
                            help='use the provided mongodb address rather than '
                            'the one listed in the config (e.g. '
                            'mongodb://192.168.1.27:27017/)')
    arg_parser.add_argument('-q', '--quiet',
                            action='store_true',
                            help='disable logging')
    arg_parser.add_argument('-v', '--verbosity',
                            default='info',
                            help='the verbosity of logging - options are: '
                            'debug, info, warning, error')
    cmd_args = arg_parser.parse_args()

    # configure logging behavior
    if not cmd_args.quiet:
        if cmd_args.verbosity.lower() == 'debug':
            log_level = logging.DEBUG
        elif cmd_args.verbosity.lower() == 'info':
            log_level = logging.INFO
        elif cmd_args.verbosity.lower() == 'warning':
            log_level = logging.WARNING
        elif cmd_args.verbosity.lower() == 'error':
            log_level = logging.ERROR
        else:
            raise Exception('didn\'t recognize logging level [%s]' \
                                        % (cmd_args.verbosity)) from None

        logging.basicConfig(level=log_level,
                        format='%(asctime)s -- %(levelname)s -- %(message)s',
                        handlers=[logging.FileHandler(cmd_args.log, 'w+'),
                                logging.StreamHandler()])

    logging.info('starting NSpyre...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    main_window = NSpyreMainWindow()
    sys.exit(app.exec())
